/*
	SMTP Transparent Proxy
	Bartlomiej Korupczynski
	(c) Warszawa 2004-2005
	GNU GPL License
	http://smtp-proxy.klolik.org

   Uzywasz tego oprogramowania na wlasne ryzyko. Autor nie ponosi
   odpowiedzialnosci za jakiekolwiek straty badz szkody spowodowane jego
   uzyciem, ani odpowiedzialnosci z jakichkolwiek innych przyczyn.
*/

/*
 *	configuration
*/

#define VERSION			"1.4.11"

/*
 * 	headers
*/

#define _GNU_SOURCE

#define _SMTP_GATED_C_
#include "setup.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <locale.h>

#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>
#include <assert.h>
//#include <err.h>

#ifdef USE_NETFILTER
#include <linux/netfilter_ipv4.h>
#endif

#ifdef USE_PGSQL
#include <libpq-fe.h>
#error USE_PGSQL not yet implemented
#endif

#ifdef USE_REGEX
#error USE_REGEX not yet implemented
#endif

#include "config.h"
#include "smtp-gated.h"
#include "lang.h"
#include "conffile.h"
#include "scan.h"
#include "util.h"
#include "daemon.h"
#include "md5.h"


/*
 * 	constants
*/

#define CONN_REJ_ER		(struct response_code) {CONN_REJ_CODE,3,0}

#ifdef USE_SHARED_MEM
#define SHARED_CONN_STATUS(x, y)	connections[child_slot].x = (y);
#define SHARED_STATS_INC(x) stats->x++;
#else
#define SHARED_CONN_STATUS(x, y)
#define SHARED_STATS_INC(x)
#endif



//#define NO_FORK
static const int one = 1;

// max. 8 chars
char *conn_states[] = { "start", "ident", "connect", "pre", "data", "bdat",
	"direct", "scan", "scan1", "scan2", "spam", "spam1", "spam2",
	"post", "quit" };


/*
	global variables
*/

char *config_file = NULL;
int spool_max_size = 0;

int child_status = 0;
int children = 0;
int i_am_a_child = 0;	

volatile sig_atomic_t force_reconfig = 0;
volatile sig_atomic_t force_finish = 0;
volatile sig_atomic_t force_dump = 0;
volatile sig_atomic_t child_died = 0;
volatile sig_atomic_t timedout = 0;

#ifdef USE_SHARED_MEM
// child_slot i i_am_a_child sie poniekad dubluja
#define I_AM_A_CHILD()	(child_slot != -1)
int child_slot = -1;
int conn_shmid = -1;
int stat_shmid = -1;
#endif

struct conn_info *connections = NULL;
struct stat_info *stats = NULL;

int max_connections_real;



/*
 * 	configuration variables
*/

char *msg_success = "Success / Sukces";


char **lookup_errors[] = {
	&msg_success,		// LOOKUP_OK
	&msg_lookup_timeout,	// LOOKUP_TIMEOUT
	&msg_lookup_mismatch,	// LOOKUP_MISMATCH
	&msg_lookup_nomem	// LOOKUP_NOMEM
};


/*
 * 	forward functions
*/

void cleanup();
int lockfile_touch(struct proc_data *data, char *vir_name);
void some_action(struct proc_data *data);



/*
 * 	deklaracje funkcji wprzod
*/

void wait_for_quit(struct proc_data *data, char* format, ...);

/*
 * 	funkcje pomocnicze
*/

inline int min(int a, int b)
{
	return (a < b) ? a : b;
}


inline void upcase(char *str)
{
	if (str == NULL) return;

	for (; str[0] != '\0'; str++) {
		if (str[0] >= 'a' && str[0] <= 'z') str[0] -= ('a'-'A');
	}
}

inline int resp_code(char *s)
{
	if (s[0]<'0' || s[0]>'9' || s[1]<'0' || s[1]>'9' || s[2]<'0' || s[2]>'9') return -1;

	return 100*((int) (s[0]-'0')) + 10*((int) (s[1]-'0')) + ((int) (s[2]-'0'));
}



/*
 * 	Logowanie
*/

#ifdef USE_SHARED_MEM
void set_dump_state(conn_state s)
{
	connections[child_slot].state = s;
}
#endif


/*
	return:
	=0	success
	<0	errno
	1	data truncated
*/

// wyslij odpowiedz code " " [ class "." subject "." detail ]
// code = tradycyjny kod odpowiedzi
// esc2 = subject kodu rozszerzonego
// esc3 = detail kodu rozszerzonego
//
#define ER(generic,subject,detail) (struct response_code) {generic, subject, detail}

int response(struct proc_data *data, struct response_code resp, char *format, ...)
	__attribute__ ((format (printf, 3, 4)));


int response(struct proc_data *data, struct response_code resp, char *format, ...)
{
	int class;
	va_list ap;
	char buf[PRINTF_SIZE];
	int res;

	class = resp.generic / 100;

	if (data->enhancedstatuscodes && (class == 2 || class == 4 || class == 5)) {
		// kod rozszerzony
		res = fdprintf(data->client, "%03d %d.%d.%d ", resp.generic, class, resp.subject, resp.detail);
	} else {
		// kod zwykly
		res = fdprintf(data->client, "%03d ", resp.generic);
	}

	if (res == -1) return -1;
	if (!format) return 0;

	va_start(ap, format);
	res = vsnprintf(buf, sizeof(buf), format, ap);
	va_end(ap);

	if (res == -1) {
		errno = ENOMEM;
		return -1;
	}

	for (;;) {
		if (timedout) {
			CLEAR_TIMEOUT();
			errno = ETIMEDOUT;
			return -1;
		}
		
	       	if ((res = write(data->client, buf, res)) == -1) {
//			if (errno == EINTR) continue;
			return -1;
		}

		if (res == 0) {
			errno = ETIMEDOUT;
			return -1;
		}

		break;
	}

	return res;
} /* response() */

/*
 * 	usuwanie i przywracanie CRLF
*/

inline void save_remove_crlf(char *buffer, char *pos, char **replace_pos, char *replace_char)
{
	if ((pos > buffer) && (*(pos-1) == '\r')) {
		*replace_pos = pos-1;
		*replace_char = **replace_pos;
		**replace_pos = '\0';
	} else {
		*replace_pos = pos;
		*replace_char = **replace_pos;
		**replace_pos = '\0';
	}
} /* rave_remove_crlf() */

inline void restore_crlf(/*@unused@*/ char *buffer, /*@unused@*/ char *pos, char **replace_pos, char *replace_char)
{
	**replace_pos = *replace_char;
} /* restore_crlf() */

/*
 * 	pipeline - kolejkowanie
*/

int queue_commandp(smtp_command command, struct proc_data *data, command_parm_t parm1, command_parm_t parm2)
{
	struct command_entry *ce;
	
	assert(data->command_pos_client < PIPELINE_QUEUE);

//	action(LOG_DEBUG, "QUEUE:QUEUE %d at %d", command, data->command_pos_client);
	ce = &data->commands[data->command_pos_client];
	ce->cmd = command;
	ce->parm1 = parm1;
	ce->parm2 = parm2;

	data->command_pos_client++;
	data->command_pos_client %= PIPELINE_QUEUE;

	// queue full
	if (data->command_pos_client == data->command_pos_server) return -1;
	return 0;
} /* queue_commandp() */

inline int queue_command(smtp_command command, struct proc_data *data)
{
	return queue_commandp(command, data, 0, 0);
} /* queue_command() */

smtp_command poll_commandp(struct proc_data *data, command_parm_t *parm1, command_parm_t *parm2)
{
	smtp_command command;
	struct command_entry *ce;

	assert(data->command_pos_server < PIPELINE_QUEUE);

	if (data->command_pos_client == data->command_pos_server) {
//		action(LOG_DEBUG, "QUEUE:POLL no commands queued");
		return COMMAND_NONE;
	}

	ce = &data->commands[data->command_pos_server];
	command = ce->cmd;
	if (parm1) *parm1 = ce->parm1;
	if (parm2) *parm2 = ce->parm2;

//	action(LOG_DEBUG, "QUEUE:POLL %d at %d", command, data->command_pos_server);
	return command;
} /* poll_commandp() */

inline void dequeue_command(struct proc_data *data)
{
	// move to next
	data->command_pos_server++;
	data->command_pos_server %= PIPELINE_QUEUE;
} /* dequeue_command() */


/*
 * 	signals
*/

void child_reaper()
{
	pid_t res;
	int i;

	for (;;) {
		res = wait3((int *) &child_status, WNOHANG, NULL);

		// brak kolejnych bachorow (teoretycznie)
		if (res == 0) break;

		// error
		if (res == -1) {
			if (errno == ECHILD) break;
			if (errno == EINTR) continue;
			
			action(LOG_ERR, "!ERR! wait3() returned: %s", strerror(errno));
			break;
		}

		if (WIFSIGNALED(child_status)) {
			action(LOG_ALERT, "!BUG! Child %d exited on signal %d", res, WTERMSIG(child_status));
		}

		if (!i_am_a_child && connections) {
			for (i=0; i<max_connections_real; i++) {
				if (connections[i].pid != res) continue;
				connections[i].pid = 0;
#ifdef USE_SHARED_MEM
				connections[i].ident_ok = 0;
#endif
			}
		}

		if (children > 0) children--;
	}
} /* child_reaper() */

void got_signal(int sig)
{
	if (sig == SIGCHLD) {
		child_died = 1;
		return;
	} else if (sig == SIGALRM) {
		timedout = 1;
		return;
	} else if (sig == SIGHUP) {
		force_reconfig = 1;
		return;
	} else if (sig == SIGUSR1) {
		force_dump = 1;
		return;
	} else if (sig == SIGUSR2) {
		return;
	} else if (sig == SIGTERM) {
		force_finish = 1;
	} else {
		action(LOG_INFO, "Received signal %s [%d], exiting.", strsignal(sig), sig);
		cleanup();
		exit(0);
	}
} /* got_signal() */

void cleanup()
{
	if (!i_am_a_child) {
		remove_pidfile();
#ifdef USE_SHARED_MEM
		shmfreeid(conn_shmid);
		shmfreeid(stat_shmid);
	} else {
		if (child_slot != -1) {
			SHARED_CONN_STATUS(ident_ok, 0);
		}
#endif
	}
} /* cleanup() */

int setup_signal()
{
	struct sigaction sa;

	sa.sa_handler = &got_signal;
	sigfillset(&sa.sa_mask);
	sa.sa_flags = 0;

	if (sigaction(SIGHUP, &sa, NULL) != 0) return -1;
	if (sigaction(SIGINT, &sa, NULL) != 0) return -1;
	if (sigaction(SIGQUIT, &sa, NULL) != 0) return -1;
	if (sigaction(SIGUSR1, &sa, NULL) != 0) return -1;
	if (sigaction(SIGUSR2, &sa, NULL) != 0) return -1;

	if (sigaction(SIGTERM, &sa, NULL) != 0) return -1;
	if (sigaction(SIGALRM, &sa, NULL) != 0) return -1;
	if (sigaction(SIGCHLD, &sa, NULL) != 0) return -1;

	sa.sa_handler = SIG_IGN;

	if (sigaction(SIGPIPE, &sa, NULL) !=0) return -1;

	return 0;
} /* setup_signal() */



/*
 * 	line reading
*/


// TODO flush_buffers()
// TODO: wysle drugi raz np. BDAT ...
// mysi pominac to co juz trafilo do func()
void flush_buffers(struct proc_data *data)
{
	ssize_t res;
	char *buf;

	buf = data->cli_buf;
	while (data->cli_size > 0) {
		res = write(data->server, buf, data->cli_size);
		if (res == -1) {
			if (errno == EINTR) continue;
			break;
		}

		data->cli_size -= res;
		buf += res;
	}

	buf = data->srv_buf;
	while (data->srv_size > 0) {
		res = write(data->client, buf, data->srv_size);
		if (res == -1) {
			if (errno == EINTR) continue;
			break;
		}

		data->srv_size -= res;
		buf += res;
	}
} /* flush_buffers() */



inline void helo(int client)
{
	fdprintf(client, "220 %s SMTP\r\n", proxy_name);
} /* helo() */

/*
 * oczekiwanie na QUIT
 * symuluje sesje SMTP, w oczekiwaniu na QUIT klienta
*/

line_status local_callback(char* buffer, char* pos, /*@unused@*/ int size, void *ptr)
{
	char *replace_pos, replace_char;
	struct proc_data *data = ptr;

	save_remove_crlf(buffer, pos, &replace_pos, &replace_char);

	if (foreground) printf("<=  %s\n", buffer);

	// close after couple of meaningless commands
	data->command_count++;
	if (data->command_count > 10) {
		fdprintf(data->client, "554 %s %s\r\n", proxy_name, msg_transaction_failed);
//		response(data, R(554,5,0), "%s %s\r\n", proxy_name, msg_transaction_failed);
		return LINE_CLOSED;
	}

	if (strcasecmp(buffer, "QUIT") == 0) {
		fdprintf(data->client, "221 %s %s.\r\n", proxy_name, msg_sign_off);
		return LINE_CLOSED;
	} else if ((strcasecmp(buffer, "NOOP") == 0) || 
		   (strcasecmp(buffer, "RSET") == 0)) {
		fdprintf(data->client, "250 OK\r\n");
	} else if ((strncasecmp(buffer, "EHLO ", 5) == 0) ||
		   (strncasecmp(buffer, "HELO ", 5) == 0)) {
		fdprintf(data->client, "250 %s %s\r\n", proxy_name, msg_hello);
	} else if (((strncasecmp(buffer, "RCPT ", 5) == 0)) ||
		   ((strcasecmp(buffer, "DATA") == 0))) {
		fdprintf(data->client, "503 %s %s\r\n", proxy_name, msg_proto_error);
	} else if ((strncasecmp(buffer, "MAIL ", 5) == 0)) {
		fdprintf(data->client, "451 %s %s\r\n", proxy_name, data->message);
	} else {
		fdprintf(data->client, "502 %s %s\r\n", proxy_name, msg_unimpl_command);
	}

	return LINE_OK;
} /* local_callback() */

// bez \r\n na koncu!

void wait_for_quit(struct proc_data* data, char* format, ...)
	__attribute__ ((format (printf, 2, 3)));
	
void wait_for_quit(struct proc_data* data, char* format, ...)
{
	va_list ap;
	char cli_buf[256];
	int cli_size, res;

	// inicjacja unii dla wait_for_quit
	data->command_count = 0;

	va_start(ap, format);
	if (vasprintf(&data->message, format, ap) == -1) {
		data->message = NULL;
	}
	va_end(ap);

	SHARED_CONN_STATUS(state, CONN_QUIT);
	action(LOG_WARNING, "SESSION TAKEOVER: host=%s, ident=%s, trns=%d, reason=%s",
		data->origin_str, data->ident, data->transaction, data->message);

	cli_size = 0;
	SET_TIMEOUT(timeout_session);

	for (;;) {
		assert(cli_size < sizeof(cli_buf));
		if (timedout) {
			CLEAR_TIMEOUT();
			res = LINE_EINTR;
			response(data, ER(421,5,0), "%s %s\r\n", proxy_name, msg_session_timeout);
			action(LOG_INFO, "CLOSE:TAKEN %s", msg_session_timeout);
			break;
		}

		res = fdgetline(data->client, cli_buf, sizeof(cli_buf), &cli_size, &local_callback, data, NULL);
		if (res == LINE_EINTR) {
			CLEAR_TIMEOUT();
			response(data, ER(421,5,0), "%s %s\r\n", proxy_name, msg_session_timeout);
			action(LOG_INFO, "CLOSE:TAKEN %s", msg_session_timeout);
			break;
		}
		if (res == LINE_CLOSED) break;
	}

	SAFE_CLOSE(data->client);
	cleanup();
	
	if (res != LINE_EINTR) action(LOG_DEBUG, "CLOSE:TAKEN");
	exit(5);
} /* wait_for_quit() */


/*
 * polaczenie pass-thru
 * zestawia "tunel" pomiedzy serwerem i klientem, nie ingeruje w przesylane dane.
*/

void direct_proxy(struct proc_data *data)
{
	fd_set rfds;
	struct timeval tv;
	int max_fd, size;
	int res;
	char buffer[buffer_size];


	CLEAR_TIMEOUT();

	// deskryptor dla select-a
	max_fd = (data->client > data->server) ? data->client : data->server;

	SHARED_STATS_INC(requests_direct);
	SHARED_CONN_STATUS(state, CONN_DIRECT);

	for (;;) {
		FD_ZERO(&rfds);
		FD_SET(data->client, &rfds);
		FD_SET(data->server, &rfds);
		tv.tv_sec = timeout_direct;
		tv.tv_usec = 0;

		if ((res = select(max_fd+1, &rfds, NULL, NULL, &tv)) == -1) {
			action(LOG_ERR, "select: %s", strerror(errno));
			continue;
		}

		// timeout
		if (res == 0) break;

		// dane od klienta
		if (FD_ISSET(data->client, &rfds)) {
			if ((size = read(data->client, buffer, sizeof(buffer))) == -1) {
				action(LOG_ERR, "direct:client:read error: %s", strerror(errno));
				break;
			} else if (size == 0) {
				break;
			} else {
				data->cli_rx += size;
				SHARED_CONN_STATUS(cli_rx, data->cli_rx);
				write(data->server, buffer, size);
			}
		}

		// dane z serwera
		if (FD_ISSET(data->server, &rfds)) {
			if ((size = read(data->server, buffer, sizeof(buffer))) == -1) {
				action(LOG_ERR, "direct:server:read error: %s", strerror(errno));
				break;
			} else if (size == 0) {
				break;
			} else {
				data->srv_rx += size;
				SHARED_CONN_STATUS(srv_rx, data->srv_rx);
				write(data->client, buffer, size);
			}
		}

	}

	SAFE_CLOSE(data->client);
	SAFE_CLOSE(data->server);

	cleanup();
	
	action(LOG_NOTICE|LOG_ALWAYS, "CLOSE:DIRECT rcv=%d/%d, time=%ld, host=%s, ident=%s",
		data->cli_rx, data->srv_rx, time(NULL)-data->time, data->origin_str, data->ident);
	exit(0);
} /* direct_proxy */


/*
	identyfikacja adres serwera oryginalnego
	=0	ok
	-1	error, (->errno)
	>0	LOOKUP_x
*/

#ifdef USE_NETFILTER
int netfilter_lookup(struct proc_data *data)
{
	int size;

	data->target.sin_family = AF_INET;
	size = sizeof(data->target);

	if (getsockopt(data->client, SOL_IP, SO_ORIGINAL_DST, &data->target, &size))
		return -1;

	snprintf(data->target_str, sizeof(data->target_str), "%s", inet_ntoa(data->target.sin_addr));
	TERMINATE_STRING(data->target_str);
	data->target_port = ntohs(data->target.sin_port);

	return 0;
} /* netfilter_lookup() */
#endif

int ident_lookup(struct proc_data *data, struct sockaddr_in host, int local_port, struct sockaddr_in local_addr)
//, socklen_t sizeof_local_addr)
{
	unsigned char buf[1024];
	char ident_tmp[IDENT_SIZE+1];	// fixed 128!
	int res, client_port, pos;
	int rport, lport, ip1, ip2, ip3, ip4;
	int sock;

	client_port = ntohs(host.sin_port);
	SET_TIMEOUT(timeout_lookup);

	if ((sock = connect_host(host, lookup_port, local_addr)) == -1) {
		goto fail;
	}

	if (fdprintf(sock, "%d , %d\r\n", client_port, local_port) == -1) {
		if (errno == EINTR) errno = ETIMEDOUT;
		goto fail;
	}

	for (pos=0;;) {
		if (timedout) {
			CLEAR_TIMEOUT();
			errno = ETIMEDOUT;
			goto fail;
		}
		if ((res = read(sock, buf+pos, sizeof(buf)-pos)) == -1) {
			goto fail;
		}
		if (res == 0) break;
		pos += res;
	}

	buf[(pos < sizeof(buf)) ? pos : sizeof(buf)-1] = '\0';

	SAFE_CLOSE(sock);
	sock = -1;

#if IDENT_SIZE != 64
#warning Please update numbers in two places:
#warning 1. in sscanf format below (last number)
#warning 2. in #if above
#error Both of them must much IDENT_SIZE!
#endif
	res = sscanf(buf, "%d , %d : ORIGIN : %d.%d.%d.%d : %d : %64s", 
		&rport, &lport, &ip1, &ip2, &ip3, &ip4, &data->target_port, ident_tmp);

	untaint(buf, sizeof(buf));

	if (res != 8) goto lookup_mismatch;
	TERMINATE_STRING(ident_tmp);

	if ((rport != client_port) || (lport != local_port) ||
		(data->target_port < 0) || (data->target_port > 65535) ||
		(ip1 < 0) || (ip1 > 255) || (ip2 < 0) || (ip2 > 255) ||
		(ip3 < 0) || (ip3 > 255) || (ip4 < 0) || (ip3 > 255)) 
		goto lookup_mismatch;

	res = snprintf(data->target_str, sizeof(data->target_str), "%d.%d.%d.%d", ip1, ip2, ip3, ip4);
	if (res == -1) return LOOKUP_NOMEM;

	TERMINATE_STRING(data->target_str);

	data->target.sin_family = AF_INET;
	data->target.sin_port = htons(data->target_port);
	if (!inet_aton(data->target_str, &data->target.sin_addr))
		goto lookup_mismatch;

	strncpy(data->ident, ident_tmp, sizeof(data->ident));
	TERMINATE_STRING(data->ident);

	// LOOKUP_RFC, LOOKUP_EXTENDED
	return LOOKUP_OK;

lookup_mismatch:
	action(LOG_ERR, "Lookup mismatch from %s [%s]", data->origin_str, buf);
	return LOOKUP_MISMATCH;

fail:
	res = errno;
	CLEAR_TIMEOUT();
	if (sock != -1) SAFE_CLOSE(sock);

	if (res == EINTR || res == ETIMEDOUT) return LOOKUP_TIMEOUT;
	errno = res;
	return -1;
} /* ident_lookup() */



/*
 * 	funkcje obslugi spoola
*/

void spool_close(struct proc_data *data)
{
	if (data->spool_fd == -1) return;

	SAFE_CLOSE(data->spool_fd);
	data->spool_fd = -1;
} /* spool_close() */

void spool_remove(struct proc_data *data)
{
	spool_close(data);

	if (data->spool_exists) {
		unlink(data->spool_name);
		data->spool_exists = 0;
	}
} /* spool_remove() */

int spool_create(struct proc_data *data)
{
	int res;

	if (spool_max_size <= 0 || !data->spool_name) return 0;

	if (data->spool_fd != -1) SAFE_CLOSE(data->spool_fd);
	data->size = data->extra_size = 0;

	data->spool_fd = open(data->spool_name, O_WRONLY | O_CREAT | O_TRUNC, spool_perm);
	if (data->spool_fd == -1) {
//		res = errno;
		data->spool_exists = 0;
		action(LOG_ERR, "SPOOL:OPEN ERROR file=%s, error=%s",
			data->spool_name, strerror(errno));

		return (ignore_errors) ? 0 : -1;
	}

	data->spool_exists = 1;

	// data->time, pid procesu sa w nazwie pliku
	res = fdprintf(data->spool_fd, "%s: src=%s:%d, ident=%s, dst=%s:%d, trns=%d\r\n", 
		SPOOL_INFO_HEADER, data->origin_str, data->origin_port, data->ident,
		data->target_str, data->target_port, data->transaction);

	if (res == -1) {
		res = errno;
		action(LOG_ERR, "SPOOL:INFO fdprintf() error: %s", strerror(errno));
		spool_remove(data);
		errno = res;
		if (!ignore_errors) return -1;
	} else {
		data->extra_size += res;
	}

	return 0;
} /* spool_create() */

int spool_write(struct proc_data *data, void *buffer, int size)
{
	int res;

	if (!size) return 0;

	data->size += size;
	if (data->spool_fd == -1) return 0;

	while (size > 0) {
		if ((res = write(data->spool_fd, buffer, size)) == -1) {
			if (errno == EINTR) continue;
			action(LOG_ERR, "spool:write() error file=%s error=%s",
				data->spool_name, strerror(errno));
			spool_remove(data);
			return -1;
		}

		// i tak go nie przeskanujemy bo za duzy, wiec nie ma sensu dalej spoolowac
		if (data->size > spool_max_size) {
			spool_remove(data);
			break;
		}
		
		size -= res;
		buffer += res;
	}

	return 0;
} /* spool_write() */


/*
 * spool_scan(data)
 * return: NULL = OK
 *         !NULL = error message
 *
 */

char* spool_scan(struct proc_data *data)
{
	static char error[256];
	int scan_start, scan_time;
	int level;
	char *scan_result_string;
	av_result scan_res;
	spam_result spam_res;
	double spam_score, cur_load;
	int do_unlink = 1;
	char *ret = NULL;
	

	/*
	 *	antivirus scanning
	*/
	
	SHARED_CONN_STATUS(state, CONN_SCAN);
	// skanowanie, jesli jest co skanowac
	if (data->spool_exists && (scan_max_size > 0) && (data->size <= scan_max_size)) {
		action(LOG_DEBUG, "DATA:SCANNING size=%d, host=%s, ident=%s",
			data->size, data->origin_str, data->ident);
		scan_start = time(NULL);
		SET_TIMEOUT(timeout_scanner);
		scan_res = av_scanner(data->spool_name, &scan_result_string);
		CLEAR_TIMEOUT();
		scan_time = time(NULL) - scan_start;
		data->virus_name = scan_result_string;
	} else {
		scan_res = SCAN_SKIPPED;
		scan_time = -1;
	}

	if (scan_res == SCAN_OK) {
		// usun plik tymczasowy
		action(LOG_INFO, "SCAN:CLEAN size=%d, time=%d, host=%s, ident=%s",
			data->size, scan_time, data->origin_str, data->ident);
	} else if (scan_res == SCAN_SKIPPED) {
		action(LOG_INFO, "SCAN:SKIPPED size=%d, host=%s, ident=%s",
			data->size, data->origin_str, data->ident);
	} else if (scan_res == SCAN_VIRUS) {
		if (!scan_result_string) scan_result_string = msg_unknown_virus;
		if (leave_on_virus) do_unlink = 0;
		SHARED_STATS_INC(viruses);

		action(LOG_WARNING, "SCAN:VIRUS size=%d, time=%d, host=%s, ident=%s, virus=%s",
			data->size, scan_time, data->origin_str, data->ident, scan_result_string);

		snprintf(error, sizeof(error), "%s (%s)", msg_virus_found, scan_result_string);
		TERMINATE_STRING(error);
		ret = error;

		if (lock_duration) {
			// some_action tylko raz
			if (lockfile_touch(data, scan_result_string) < 1)
				some_action(data);
		} else {
			some_action(data);
		}
	} else if (scan_res == SCAN_TIMEOUT) {
		if (leave_on_error) do_unlink = 0;
		action(LOG_WARNING, "SCAN:TIMEOUT size=%d, time=%d, host=%s, ident=%s, result=%s",
			data->size, scan_time, data->origin_str, data->ident, scan_result_string);
		if (!ignore_errors) {
			snprintf(error, sizeof(error), "%s", msg_scanner_failed);
			TERMINATE_STRING(error);
			ret = error;
		}
	} else {
		// usun plik tymczasowy
		if (leave_on_error) do_unlink = 0;
		action(LOG_ERR, "SCAN:FAILED size=%d, time=%d, host=%s, ident=%s, result=%s",
			data->size, scan_time, data->origin_str, data->ident, scan_result_string);
		if (!ignore_errors) {
			snprintf(error, sizeof(error), "%s", msg_scanner_failed);
			TERMINATE_STRING(error);
			ret = error;
		}
	}

	/*
	 *	antiSPAM scanning
	*/

	if ((spam_max_size > 0) && (ret == NULL)) {
		SHARED_CONN_STATUS(state, CONN_SPAM);
		if (is_load_above(spam_max_load, &cur_load) == 1) {
			action(LOG_INFO, "SPAM:SKIPPED size=%d, host=%s, ident=%s, load=%lf",
				data->size, data->origin_str, data->ident, cur_load);
		} else if (data->spool_exists && (data->size <= spam_max_size)) {
			spam_score = 0.0;
	
			scan_start = time(NULL);
			SET_TIMEOUT(timeout_spam);
			spam_res = spam_scanner(data->spool_name, &spam_score);
			CLEAR_TIMEOUT();
			scan_time = time(NULL) - scan_start;

			if (spam_res == SPAM_OK) {
				spam_res = (spam_score >= spam_threshold) ? SPAM_YES : SPAM_NO;
			}

			if (spam_res == SPAM_YES) {
				scan_result_string = "FOUND";
				level = LOG_NOTICE;
				if (leave_on_spam) do_unlink = 0;
				SHARED_STATS_INC(spams);

				if (spam_block) {
					snprintf(error, sizeof(error), "%s [%lf]", msg_spam_found, spam_score);
					TERMINATE_STRING(error);
					ret = error;
				}
			} else if (spam_res == SPAM_NO) {
				scan_result_string = "CLEAN";
				level = LOG_INFO;
			} else if (spam_res == SPAM_TIMEOUT) {
				scan_result_string = "TIMEOUT";
				level = LOG_WARNING;
				if (leave_on_error) do_unlink = 0;
			} else {
				scan_result_string = "FAILED";
				level = LOG_ERR;
				if (leave_on_error) do_unlink = 0;
			}
	
			action(level, "SPAM:%s size=%d, time=%d, host=%s, ident=%s, result=%lf",
				scan_result_string, data->size, scan_time, data->origin_str,
				data->ident, spam_score);
		} else {
			action(LOG_INFO, "SPAM:SKIPPED size=%d, host=%s, ident=%s",
				data->size, data->origin_str, data->ident);
		}
	}

	if (do_unlink) spool_remove(data);
	return ret;
} /* spool_scan() */


/*
 * 	logowanie adresow z naglowka
 *	zakladamy: line w postaci zakonczonej '\0'
*/

// MAIL FROM: bartek@isp.pl
// MAIL FROM: <bartek@isp.pl>
// MAIL FROM:<bartek@isp.pl   > BODY=8BIT
char* alloc_msg_mail(char *line, int len)
{
	char buf[len+1];
	register unsigned char c;
	int i;

	if (!line) return NULL;

	// pomin spacje i tabulatory i '<' na poczatku
	while ((line[0] == ' ') || (line[0] == '\t') || (line[0] == '<')) line++;

	for (i=0; i<sizeof(buf); i++) {
		c = line[i];

		switch (c) {
		case '>':
		case ' ':
		case '\0':
		case '\t':
			buf[i] = '\0';
			goto quit;
		default: 
			/* do not untaint here, just copy */
			buf[i] = c;
		}
	}

quit:
	TERMINATE_STRING(buf);

	return strdup(buf);
} /* alloc_msg_mail() */


char* alloc_str_crlf(char *line, int len)
{
	char buf[len+1];
	register unsigned char c;
	int i;

	if (!line) return NULL;

	for (i=0; i<sizeof(buf); i++) {
		c = line[i];

		switch (c) {
		case '\r':
		case '\n':
			buf[i] = '#';
			break;
		case '\0':
			buf[i] = '\0';
			goto quit;
		default:
			/* do not untaint here, just copy */
			buf[i] = c;
		}
	}
	
quit:
	TERMINATE_STRING(buf);
	
	return strdup(buf);
}

/*
 *	naglowek wiadomosci
*/

void prepare_nat_header(struct proc_data *data)
{
//	char *buf_from, *buf_ident, *buf_with, *buf_abuse;
	char buf_from[64], buf_ident[16+IDENT_SIZE], buf_with[128], buf_abuse[64];

	snprintf(buf_from, sizeof(buf_from), "%s: from %s ", HEADER, inet_ntoa(data->origin.sin_addr));

	if (!EMPTY_STRING(data->ident)) {
		snprintf(buf_ident, sizeof(buf_ident), "[ident %s]\r\n", data->ident);
	} else {
		snprintf(buf_ident, sizeof(buf_ident), "[ident-empty]\r\n");
	}

	snprintf(buf_with, sizeof(buf_with), "\tby %s with TPROXY id %ld.%d\r\n",
		proxy_name, data->time, getpid());

	if (!EMPTY_STRING(abuse)) {
		snprintf(buf_abuse, sizeof(buf_abuse), "\tabuse-to %s\r\n", abuse);
	} else {
		buf_abuse[0] = '\0';
	}

	TERMINATE_STRING(buf_from);
	TERMINATE_STRING(buf_ident);
	TERMINATE_STRING(buf_with);
	TERMINATE_STRING(buf_abuse);

	data->header_size = asprintf(&data->header, "%s%s%s%s", buf_from, buf_ident, buf_with, buf_abuse);
	if (data->header_size == -1) {
		data->header = NULL;
		data->header_size = 0;
	}
} /* prepare_nat_header() */


/*
 * 	obsluga logowania mail from/rcpt to
*/

void destroy_addresses(struct proc_data *data)
{
	int i;

	if (data->mail_from) {
		free(data->mail_from);
		data->mail_from = NULL;
	}

	for (i=0; i<RCPTS_ONE_TIME; i++) {
		if (!data->rcpt_to[i]) continue;

		free(data->rcpt_to[i]);
		data->rcpt_to[i] = NULL;
	}
	data->rcpts = 0;
} /* destroy_addresses() */

void flush_addresses(struct proc_data *data)
{
	if (!data->mail_from && !data->rcpts) return;

	if (data->mail_from && !data->rcpts) {
		action(LOG_INFO|LOG_ALWAYS, "MAIL FROM <%s>", data->mail_from);
		free(data->mail_from);
		data->mail_from = NULL;
		return;
	}

	// teoretycznie niepotrzebne
	if (!data->rcpts) return;

	if (data->mail_from) {
		action(LOG_INFO|LOG_ALWAYS, "MAIL FROM <%s> RCPT TO: %03d<%s>", 
			data->mail_from, data->rcpt_to_code[0], data->rcpt_to[0]);
		free(data->mail_from);
		data->mail_from = NULL;
		
		free(data->rcpt_to[0]);
		data->rcpt_to[0] = data->rcpt_to[1];
		data->rcpt_to_code[0] = data->rcpt_to_code[1];
		data->rcpt_to[1] = NULL;
		
		data->rcpts--;
		return;
	}

	// data->rcpts > 0
	if (data->rcpts == 1) {
		action(LOG_INFO|LOG_ALWAYS, "RCPT TO: %03d<%s>",
			data->rcpt_to_code[0], data->rcpt_to[0]);
	} else {
		action(LOG_INFO|LOG_ALWAYS, "RCPT TO: %03d<%s>, %03d<%s>",
			data->rcpt_to_code[0], data->rcpt_to[0],
			data->rcpt_to_code[1], data->rcpt_to[1]);
	}
	destroy_addresses(data);
} /* flush_addresses() */

// email is malloc-ed
void new_mail_from(struct proc_data *data, char *email, int code)
{
	char *tmp;
	int accepted = (code == 250);
	
	if ((log_mail_from & (accepted ? LOG_MAIL_ACCEPTED : LOG_MAIL_REJECTED)) == 0)
		return;

	if (log_mail_from & LOG_MAIL_BASE64) {
		tmp = strdup(md5_string_base64(email));
		free(email);
		email = tmp;
	} else {
		untaint(email, -1);
	}

	if (!accepted) {
		action(LOG_NOTICE|LOG_ALWAYS, "MAIL FROM <%s> rejected [%d]", email, code);
		free(email);
		return;
	}
	
	if (data->mail_from) free(data->mail_from);
	data->mail_from = email;
} /* net_mail_from() */

// email is malloc-ed
void add_rcpt_to(struct proc_data *data, char *email, int code)
{
	char *tmp;

	if ((log_rcpt_to & (code == 250 ? LOG_MAIL_ACCEPTED : LOG_MAIL_REJECTED)) == 0)
		return;
	
	if (log_rcpt_to & LOG_MAIL_BASE64) {
		tmp = strdup(md5_string_base64(email));
		free(email);
		email = tmp;
	} else {
		untaint(email, -1);
	}

	data->rcpt_to[data->rcpts] = email;
	data->rcpt_to_code[data->rcpts] = code;
	data->rcpts++;

	if ((data->rcpts >= RCPTS_ONE_TIME) || data->mail_from) flush_addresses(data);
} /* add_rcpt_to() */


/*
 * 	funkcje obslugi sesji
*/

void transaction_reset(struct proc_data *data)
{
	data->data_used = 0;
	data->bdat_used = 0;
	data->bdat_togo = 0;
	data->bdat_last = 0;

	data->size = 0;
	data->extra_size = 0;
	data->rcpts = 0;

	destroy_addresses(data);
} /* transaction_reset() */

line_status client_callback(char *buffer, char *pos, int size, void *ptr)
{
	struct proc_data *data = ptr;
	char *replace_pos, replace_char;
	char *msg, tmp[32];
	int p1, p2, res;
	int replaced = 0;
	line_status ret_code = LINE_OK;

	// nie zakonczona linia => nic sie nie da zrobic
	if (!pos) goto write_ok;

	save_remove_crlf(buffer, pos, &replace_pos, &replace_char);

	if (data->data_going == GOING_NONE) {
//		if (foreground) printf("<-  %s\n", buffer);
		if (strcasecmp(buffer, "DATA") == 0 || strncasecmp(buffer, "DATA ", 5) == 0) {
			// zostanie odrzucony przez serwer
			// if (data->bdat_used) {}
			action(LOG_DEBUG, "DATA:REQUEST");
			if (spool_create(data) != 0) {
				res = errno;
				fdprintf(data->server, "QUIT\r\n");
				SAFE_CLOSE(data->server);

				response(data, ER(451,3,1), "%s %s: %s\r\n", proxy_name, msg_spool_open_fail, strerror(res));
				wait_for_quit(data, "%s %s: %s", proxy_name, msg_spool_open_fail, strerror(res));
				exit(5);
			}
			// possibly receives "go ahead" [354] response
			res = queue_command(COMMAND_DATA, data);
			// pseudo-command, receives transaction ack [250]
			if (!res) res = queue_command(COMMAND_DATA_ACK, data);
		} else if (strncasecmp(buffer, "BDAT ", 5) == 0) {
			strncpy(tmp, buffer, sizeof(tmp));
			TERMINATE_STRING(tmp);
			upcase(tmp);
			// unsigned int
			if (sscanf(tmp, "BDAT %u LAST", &p1) == 1) {
				p2 = 1;
			} else if (sscanf(tmp, "BDAT %d", &p1) == 1) {
				p2 = 0;
			} else {
				p1 = p2 = 0;
			}

			if (data->chunking) {
				action(LOG_DEBUG, "BDAT:REQUEST %d %s", p1, p2 ? "LAST" : "");
				// TODO: 
				// "BDAT p1 LAST" => "BDAT p1", "BDAT 0 LAST"/"QUIT"
				// replaced = 1;
				data->bdat_togo = p1;
				data->bdat_last = p2;

				if (p2) {
					fdprintf(data->server, "BDAT %u\r\n", p1);
					replaced = 1;
//					res = queue_commandp(COMMAND_BDAT_PRELAST, data, p1, p2);
//					res = queue_commandp(COMMAND_BDAT_LAST, data, p1, p2);
					res = queue_commandp(COMMAND_BDAT, data, p1, p2);
				} else {
					res = queue_commandp(COMMAND_BDAT, data, p1, p2);
				}
			} else {
				action(LOG_DEBUG, "BDAT without CHUNKING support host=%s, ident=%s",
					data->origin_str, data->ident);

				buffer[3] = '*';
//				response(data, ER(500,5,1), "[%s] Command used when not advertised\r\n", proxy_name);
//				replaced = 1;
				res = queue_command(COMMAND_OTHER, data);
			}
			ret_code = LINE_BINARY;
		} else if (strcasecmp(buffer, "STARTTLS") == 0) {
			action(LOG_DEBUG, "STARTTLS:REQUEST");
			res = queue_command(COMMAND_STARTTLS, data);
		} else if (strcasecmp(buffer, "RSET") == 0) {
			res = queue_command(COMMAND_RSET, data);
		} else if (strcasecmp(buffer, "QUIT") == 0) {
			res = queue_command(COMMAND_QUIT, data);
		} else if (strncasecmp(buffer, "RCPT TO:", 8) == 0) {
			res = queue_commandp(COMMAND_RCPT, data, (int) alloc_msg_mail(buffer+8, EMAIL_LENGTH), 0);
		} else if (strncasecmp(buffer, "MAIL FROM:", 10) == 0) {
			res = queue_commandp(COMMAND_MAIL, data, (int) alloc_msg_mail(buffer+10, EMAIL_LENGTH), 0);
		} else if (strncasecmp(buffer, "HELO ", 5) == 0) {
			// TODO: bez \r\n
			res = queue_commandp(COMMAND_HELO, data, (int) strndup(buffer+5, HELO_LENGTH), 0);
		} else if (strncasecmp(buffer, "EHLO ", 5) == 0) {
			// TODO: bez \r\n
			res = queue_commandp(COMMAND_EHLO, data, (int) strndup(buffer+5, HELO_LENGTH), 0);
		} else {
			res = queue_command(COMMAND_OTHER, data);
		}

		if (res != 0) {
			fdprintf(data->server, "QUIT\r\n");
			SAFE_CLOSE(data->server);
			data->server = -1;

			spool_close(data);
			if (!leave_on_error) spool_remove(data);

			// TODO: a moze od razu rozlaczyc?
			response(data, ER(503,5,0), "%s %s\r\n", proxy_name, msg_pipeline_full);
			wait_for_quit(data, "%s", msg_pipeline_full);
			exit(5);
		}
	} else {
		if (strcmp(buffer, ".") == 0) {
//			if (foreground) printf("<-  %s\n", buffer);
			spool_close(data);
			msg = spool_scan(data);
			if (!EMPTY_STRING(msg)) {
				SAFE_CLOSE(data->server);
				data->server = -1;

				response(data, ER(550,7,1), "%s %s\r\n", proxy_name, msg);
				wait_for_quit(data, "%s", msg);
				exit(5);
			}

			SHARED_CONN_STATUS(state, CONN_POST);
		} else {
			if (data->data_going == GOING_HEADER) {
				// ... header 
				if (buffer[0] == '\0') {
					// empty line => message header complete
					data->data_going = GOING_BODY;
				}
			}

//			if (foreground) printf("<#  %s\n", buffer);
			restore_crlf(buffer, pos, &replace_pos, &replace_char);
			if (buffer[0] == '.') {
				// wycinamy pierwsza kropke [channel transparency]
				spool_write(data, buffer+1, size-1);
			} else {
				spool_write(data, buffer, size);
			}
		}
	}


	// komenda usunieta ze strumienia, nie wysylaj
	if (replaced) return ret_code;

	restore_crlf(buffer, pos, &replace_pos, &replace_char);

write_ok:
	for (;;) {
		if (timedout) {
			CLEAR_TIMEOUT();
			errno = ETIMEDOUT;
			return LINE_CLOSED;
		}
		if ((res = write(data->server, buffer, size)) == -1) {
			if (errno == EINTR) continue;
#ifdef SILENT_ECONNRESET
			if (errno == ECONNRESET) return LINE_CLOSED;
#endif
			action(LOG_ERR, "client:write(%d) error: %s", size, strerror(errno));
			return LINE_CLOSED;
		}

		if (res == 0) {
			action(LOG_ERR, "client:write(%d) returned 0 (connection lost)", size);
			return LINE_CLOSED;
		}
		
		break;
	}

	return ret_code;
} /* client_callback() */

line_status server_callback(char *buffer, char *pos, int size, void *ptr)
{
	char *replace_pos, replace_char;
	struct proc_data *data = ptr;
	smtp_command command;
	command_parm_t parm1, parm2;
	int res;
	int code;		// numeric response code
	int cont;		// response continues on next line
	int direct = 0;		// enter direct proxy mode
	int ret_code = LINE_OK;	// function return code
	int ack;


	// dluga linia/krotka linia => nie ma co sprawdzac
	// jesli nieprawidlowa odpowiedz z serwera to tez nie ma co robic
	if (!pos || size<3 || (code = resp_code(buffer)) == -1) goto write_ok;

	// status bedzie kontynuowany?
	cont = ((size > 3) && (buffer[3] == '-'));

	// co nas czeka
	command = poll_commandp(data, &parm1, &parm2);

#ifdef FILTER_CHUNKING
	// paskudny hack, length("250 CHUNKING")=12[+crlf]
	if (command == COMMAND_EHLO && size>=12) {
		save_remove_crlf(buffer, pos, &replace_pos, &replace_char);
		if (code == 250 && strcasecmp(buffer+4, "CHUNKING") == 0) {
			buffer[7] = '*';
			action(LOG_DEBUG, "ESMTP CHUNKING filtered-out dst=%s:%d",
				data->target_str, data->target_port);
		}
		restore_crlf(buffer, pos, &replace_pos, &replace_char);
	}
#else
#warning CHUNKING support possibly broken! use -DFILTER_CHUNKING to filter this extension
#endif

	// przygotowania
	save_remove_crlf(buffer, pos, &replace_pos, &replace_char);
//	if (foreground) printf(" -> %s\n", buffer);

	// process both continued and ending responses
	switch (command) {
	case COMMAND_EHLO:	// SMTP service extensions
		if (code == 250) {
			if (strcasecmp(buffer+4, "ENHANCEDSTATUSCODES") == 0) {
				data->enhancedstatuscodes = 1;
				action(LOG_DEBUG, "ESMTP ENHANCEDSTATUSCODES supported dst=%s:%d",
					data->target_str, data->target_port);
			} else if (strcasecmp(buffer+4, "CHUNKING") == 0) {
				data->chunking = 1;
				action(LOG_DEBUG, "ESMTP CHUNKING supported dst=%s:%d",
					data->target_str, data->target_port);
			}
		}
	default:
		break;
	}

	// odpowiedz nie zakonczona nas w tym momencie nie interesuje
	if (cont) goto restore_write_ok;

	// odpowiedz nie bedzie kontynuowana
	if (command != COMMAND_NONE) dequeue_command(data);
	
	// pipeline: liczenie odpowiedzi,
	// STARTTLS,DATA jest juz synchronizowane
	switch (command) {
	case COMMAND_DATA:
		if (code == 354) {
			data->data_used = 1;
			flush_addresses(data);
			data->data_going = GOING_HEADER;
			data->transaction++;
			
			SHARED_CONN_STATUS(transaction, data->transaction);
			SHARED_CONN_STATUS(state, CONN_DATA);

			if (nat_header && data->header) fdprintf(data->server, "%s", data->header);
			action(LOG_DEBUG, "DATA:GOING");
		} else {
			spool_remove(data);
			data->data_going = GOING_NONE;

			// dequeue pseudo-command COMMAND_DATA_ACK
			// there won't be any ack cause it's already failed
			dequeue_command(data);
			action(LOG_DEBUG, "DATA:CANCELLED [%d]", code);
		}
		break;

	case COMMAND_DATA_ACK:
		transaction_reset(data);
		data->data_going = GOING_NONE;
//		SHARED_CONN_STATUS(state, CONN_ACK);
		action(LOG_DEBUG, "DATA:FINISHED [%d]", code);
		break;

	case COMMAND_BDAT:
		if (code == 250) {
			flush_addresses(data);
			action(LOG_DEBUG, "BDAT:CHUNK ok [%d] %s", parm1, parm2 ? "LAST" : "");
			if (!data->bdat_used) {
				if (data->data_used) {}
				data->bdat_used = 1;
				// tymczasowe rozwiazanie
				flush_buffers(data);
				action(LOG_DEBUG, "BDAT:DIRECT");
				direct = 1;
				break;
				// TODO: inject for BDAT!
				//if (nat_header && data->header) inject_nat_header(data);
			}

			// BDAT last
			if (parm2) {


			}
		} else {
			action(LOG_DEBUG, "BDAT:CHUNK rejected [%d]", code);
			data->bdat_togo = 0;
			data->bdat_last = 0;
//`			data->bdat_used = 0;	// TODO: napewno?
		}
		break;

	case COMMAND_MAIL:
		transaction_reset(data);
		ack = (code == 250);
		new_mail_from(data, (char *) parm1, code);
		break;

	case COMMAND_RCPT:
		ack = (code == 250);
		add_rcpt_to(data, (char *) parm1, code);
		if (ack) data->rcpts_total++;
		break;

	case COMMAND_STARTTLS:
		if (code == 220) {
			spool_remove(data);
			action(LOG_DEBUG, "DIRECT:GOING");
			// direct_proxy() after write()
			direct = 1;
		} else {
			action(LOG_DEBUG, "DIRECT:CANCELLED [%d]", code);
		}
		break;

	case COMMAND_RSET:
		transaction_reset(data);
		break;

	case COMMAND_HELO:
	case COMMAND_EHLO:
		transaction_reset(data);
		if (log_helo) {
			untaint((char *) parm1, 0);
			action(LOG_INFO|LOG_ALWAYS, "%s host=%s, ident=%s, helo=%s\n",
				(command == COMMAND_EHLO) ? "EHLO" : "HELO",
				data->origin_str, data->ident, (char *) parm1);
		}
		free((char *) parm1);
		break;
	default:
		break;
	}

restore_write_ok:
	restore_crlf(buffer, pos, &replace_pos, &replace_char);

write_ok:
	for (;;) {
		if (timedout) {
			CLEAR_TIMEOUT();
			errno = ETIMEDOUT;
			return LINE_CLOSED;
		}
		if ((res = write(data->client, buffer, size)) == -1) {
			if (errno == EINTR) continue;
#ifdef SILENT_ECONNRESET
			if (errno == ECONNRESET) return LINE_CLOSED;
#endif
			action(LOG_ERR, "server:write(%d) error: %s", size, strerror(errno));
			return LINE_CLOSED;
		}
		
		if (res == 0) {
			action(LOG_ERR, "server:write(%d) returned 0 (connection lost)", size);
			return LINE_CLOSED;
		}
		
		break;
	}

	if (direct) {
		direct_proxy(data);
		exit(0);
	}
	
	return ret_code;
} /* server_callback() */


int target_lookup(struct proc_data *data)
{
	int res;
#ifdef USE_SHARED_MEM
	int i;
#endif

	res = 0;

	// lookups
	// must fill: target, target_str, ident
	if (!EMPTY_STRING(fixed_server)) {
		strncpy(data->target_str, fixed_server, sizeof(data->target_str));
		TERMINATE_STRING(data->target_str);
		data->target_port = fixed_server_port;

		data->target.sin_family = AF_INET;
		data->target.sin_port = htons(data->target_port);
		if (!inet_aton(data->target_str, &data->target.sin_addr))
			res = LOOKUP_MISMATCH;
//		snprintf(data->ident, sizeof(data->ident), "fixed");
//		TERMINATE_STRING(data->ident);
		// normalny ident?
#ifdef USE_NETFILTER
	} else if (use_netfilter) {
		res = netfilter_lookup(data);
#endif

//		snprintf(data->ident, sizeof(data->ident), "local");
//		TERMINATE_STRING(data->ident);

		// normalny ident?
	} else {
		res = ident_lookup(data, data->origin, port, data->local_addr);	//, data->local_addr_len);

#ifdef USE_SHARED_MEM
		if (res == 0) {
			data->ident_count = 0;
			untaint(data->ident, sizeof(data->ident));
			// strncpy zeruje pozostalosc stringa,
			strncpy(connections[child_slot].ident, data->ident, sizeof(connections[child_slot].ident));
			TERMINATE_STRING(connections[child_slot].ident);
			SHARED_CONN_STATUS(ident_ok, 1);

			for(i=0; i<max_connections_real; i++) {
				// if (!connections[i].pid) continue;
				if (!connections[i].ident_ok) continue;
				if (connections[i].src != SIN_TO_UINT32(data->origin.sin_addr)) continue;
				if (strcmp(connections[i].ident, data->ident) != 0) continue;

				data->ident_count++;
			}

			if (data->ident_count > max_per_ident) {
				/* no need to be nice for flooder */
				response(data, CONN_REJ_ER, "%s %s\n", proxy_name, msg_max_per_ident);
				SAFE_CLOSE(data->client);
				SHARED_STATS_INC(rejects_ident);
				action(LOG_INFO, "Rejecting (ident %d) connection from %s:%d [%s]",
					data->ident_count, data->origin_str, data->origin_port, data->ident);
				cleanup();
				exit(0);
			}
		}
#endif
	}

	if (res == -1) {
		res = errno;
		helo(data->client);
		wait_for_quit(data, "%s: %s", msg_lookup_failed, strerror(res));
		exit(5);
	} else if (res > 0) {
		helo(data->client);
		wait_for_quit(data, "%s: %s", msg_lookup_failed, *lookup_errors[res]);
		exit(5);
	}

	return 1;
} /* target_lookup() */


int target_connect(struct proc_data *data)
{
	int res;
	struct sockaddr_in src;


	SHARED_CONN_STATUS(dst, SIN_TO_UINT32(data->target.sin_addr));
	SHARED_CONN_STATUS(state, CONN_CONNECT);
	
	inet_aton(source_addr, &src.sin_addr);
	SET_TIMEOUT(timeout_connect);

	if ((data->server = connect_host(data->target, ntohs(data->target.sin_port), src)) == -1) {
		res = errno;
		helo(data->client);
		if (res == EINTR || res == ETIMEDOUT) {
			wait_for_quit(data, "%s [%s]", msg_connect_timeout, data->target_str);
			exit(5);
		} else {
			wait_for_quit(data, "%s [%s]: %s", msg_cannot_connect, data->target_str, strerror(res));
			exit(5);
		}
		exit(1);
	}

	CLEAR_TIMEOUT();
	return 0;
} /* target_connect() */


void context_init_1(struct proc_data *data, int client, struct sockaddr_in origin)
{
	memset(data, '\0', sizeof(struct proc_data));

	data->spool_fd = -1;
	data->origin = origin;
	data->client = client;
	data->server = -1;
	data->time = time(NULL);
	data->cli_buf = malloc(buffer_size);
	data->srv_buf = malloc(buffer_size);
	
	// adres zrodlowy (tekst)
	snprintf(data->origin_str, sizeof(data->origin_str), "%s", inet_ntoa(data->origin.sin_addr));
	TERMINATE_STRING(data->origin_str);
	data->origin_port = ntohs(data->origin.sin_port);

	// utworz nazwe lockfile, jesli nie trzeba do tego identa
	if (!EMPTY_STRING(fixed_server) || use_netfilter) {
		if (asprintf(&data->lockfile, "%s/%s", lock_path, data->origin_str) == -1) {
			data->lockfile = NULL;
		}
//		action(LOG_DEBUG, "LOCK:DEBUG using simple lock file [%s]", data->lockfile);
	}
} /* context_init_1() */

void context_init_2(struct proc_data *data)
{
	int res, len;

	// nazwa pliku spool-a
	if (asprintf(&data->spool_name, "%s/%lu.%u", spool_path, data->time, getpid()) == -1) {
		data->spool_name = NULL;
		if (!ignore_errors) {
			helo(data->client);
			wait_for_quit(data, "%s", msg_spool_problem);
			exit(5);
		}
	}

	// adres lokalny potrzebny do ident-a
	// data->local_addr_len;
	len = sizeof(data->local_addr);
	if (getsockname(data->client, (struct sockaddr *) &data->local_addr, &len) != 0) {
		res = errno;
		helo(data->client);
		wait_for_quit(data, "%s: %s", msg_lookup_failed, strerror(res));
		exit(5);
	}
} /* context_init_2() */


/*
 * 	lockfile
*/

int lockfile_present(struct proc_data *data)
{
	struct stat statinfo;
	char *tmp_ident;

	// lockfile bez identa? => sprawdzanie w innym miejscu
	if (EMPTY_STRING(data->lockfile)) return 0;

	// stat
	if (stat(data->lockfile, &statinfo) != 0) {
		// nie ma => ok
		if (errno == ENOENT) return 0;

		// problem innej natury
		action(LOG_CRIT, "LOCK:stat(%s): %s", data->lockfile, strerror(errno));
		return 0;
	}

	if (statinfo.st_uid == 0) {
		action(LOG_DEBUG, "LOCK:SAINT uid %d!=%d, host=%s, ident=%s [don't lock]",
			statinfo.st_uid, getuid(), data->origin_str, data->ident);
		return 0;
	}

	// jesli uplynal czas blokady to usun
	tmp_ident = EMPTY_STRING(data->ident) ? "-" : data->ident;

	if ((lock_duration == -1) || (time(NULL) - statinfo.st_mtime < lock_duration)) {
		action(LOG_DEBUG, "LOCK:LOCKED host=%s, ident=%s", data->origin_str, tmp_ident);
		return 1;
	} else {
		action(LOG_INFO, "LOCK:EXPIRED host=%s, ident=%s", data->origin_str, tmp_ident);
		unlink(data->lockfile);
		return 0;
	}
} /* lockfile_present() */

int lockfile_touch(struct proc_data *data, char *vir_name)
{
	int fd;


	if (!data->lockfile) return 0;
	
	fd = open(data->lockfile, O_CREAT|O_WRONLY|O_EXCL|O_NOCTTY|O_TRUNC, lock_perm);
	if (fd == -1) {
		if (errno == EEXIST) return 1;

		action(LOG_ERR, "LOCK:CREATE (%s): %s", data->lockfile, strerror(errno));
		return -1;
	}

	fdprintf(fd, "%s", vir_name);
	SAFE_CLOSE(fd);

	return 0;
} /* lockfile_touch() */

int lockfile_ident_present(struct proc_data *data)
{
	char buf[sizeof(data->ident)];

	// lockfile bez identa? => sprawdzanie w innym miejscu
	if (!EMPTY_STRING(data->lockfile)) return lockfile_present(data);

	// usun niewygodne znaki z identa
	memcpy(buf, data->ident, sizeof(buf));
	untaint_for_filename(buf, sizeof(buf));

	if (asprintf(&data->lockfile, "%s/%s-%s", lock_path, data->origin_str, buf) == -1) {
		action(LOG_DEBUG, "LOCK:lockfile_ident_present:asprintf() error -1!");
		data->lockfile = NULL;
	}

	// sprawdz czy klient jest zablokowany
	if (lock_duration) {
//		action(LOG_DEBUG, "LOCK:DEBUG ident lock file [%s]", data->lockfile);
		return lockfile_present(data);
	} else {
		return 0;
	}
} /* lockfile_ident_present() */


/*
 * 	zrob cos innego (TM) jesli znajdziemy wirusa :)
*/

void some_action(struct proc_data *data)
{
		// if (data->server != -1) SAFE_CLOSE(data->server)
		// SAFE_CLOSE(data->client)
		//
		// putenv...
		//
		// fork()
		// exec()
} /* some_action() */


/*
 * 	CHUNKING
*/

line_status bdat_chunk(struct proc_data *data, int allow_read)
{
	int size;
	int res;
	
	// TODO: co jesli klient da od razu "BDAT 0 LAST"?
	
	// nie mozemy zrobic od razu read(), bo moglismy tu trafic z danymi
	// pozostalymi z fdgetline() i kolejny read() moglby zablokowac
	size = min(buffer_size - data->cli_size, data->bdat_togo);
	if (allow_read && size) {
		res = read(data->client, data->cli_buf + data->cli_size, size);
		if (res == -1) {
#ifdef SILENT_ECONNRESET
			if (errno == ECONNRESET) return LINE_CLOSED;
#endif
			action(LOG_ERR, "bdat_chunk:read(%d): %s", size, strerror(errno));
			return LINE_CLOSED;
		}

		data->cli_size += res;
		data->cli_rx += res;
		// === if (res == 0)
		if (!data->cli_size) return LINE_CLOSED;	// a moze blad?
	}
	
	size = min(data->bdat_togo, data->cli_size);
	spool_write(data, data->cli_buf, size);

	for (;;) {
		if ((res = write(data->server, data->cli_buf, size)) == -1) {
			if (errno == EINTR) continue;
#ifdef SILENT_ECONNRESET
			if (errno == ECONNRESET) return LINE_CLOSED;
#endif
			action(LOG_ERR, "bdat_chunk:write(%d) error: %s", size, strerror(errno));
			return LINE_CLOSED;
		}
		
		if (res == 0) {
			action(LOG_ERR, "bdat_chunk:write(%d) returned 0 (connection lost)", size);
			return LINE_CLOSED;
		}
		
		break;
	}

	data->bdat_togo -= size;
	data->cli_size -= size;
	if (data->cli_size) memmove(data->cli_buf, data->cli_buf+size, data->cli_size);

	if (!data->bdat_togo && data->bdat_last) {
			spool_close(data);
			// TODO: scanning, BDAT 0 LAST/QUIT
	}

	return LINE_OK;
} /* bdat_chunk() */

/*
 * 	obsluga sesji
*/

void connection(struct proc_data *data)
{
	fd_set rfds;
	struct timeval tv;
	int res, max_fd;


	// pid, src filled by parent process after fork()-ing
	SHARED_CONN_STATUS(state, CONN_IDENT);
	SHARED_CONN_STATUS(cli_rx, 0);
	SHARED_CONN_STATUS(srv_rx, 0);

	context_init_2(data);

	// dane serwera docelowego
	if (!target_lookup(data)) exit(1);

	if (lockfile_ident_present(data)) {
		SHARED_STATS_INC(rejects_lock);
		fdprintf(data->client, "%03d %s %s\n", CONN_REJ_CODE, proxy_name, msg_virus_locked);
		SAFE_CLOSE(data->client);
		exit(0);
	}
	
	// log
	action(LOG_NOTICE|LOG_ALWAYS, "NEW (%d/%d) src=%s:%d, ident=%s, dst=%s:%d, id=%ld.%d", 
		children, data->ident_count, data->origin_str, data->origin_port,
		data->ident, data->target_str, data->target_port, data->time, getpid());

	// nie bedzie petli?
	// dla testow wylacz (albo: wlacz jesli !fixed)
	if (!foreground) {
		if (data->origin.sin_addr.s_addr == data->target.sin_addr.s_addr) {
			action(LOG_WARNING, "Avoiding loop, exiting.");
			SAFE_CLOSE(data->client);
			cleanup();
			exit(1);
		}
	}

	// sygnaly
	setup_signal();

	// podlaczenie do docelowego serwera SMTP
	if (target_connect(data) != 0) exit(1);

	// przygotowanie naglowka nat
	prepare_nat_header(data);

	// no real command, but we get initial 'response' from MTA
	queue_command(COMMAND_OTHER, data);

	// deskryptor dla select-a
	max_fd = (data->client > data->server) ? data->client : data->server;

	SHARED_CONN_STATUS(state, CONN_PRE);

	// petla obslugi sesji
	for (;;) {
		FD_ZERO(&rfds);
		FD_SET(data->client, &rfds);
		FD_SET(data->server, &rfds);
		tv.tv_sec = timeout_idle;
		tv.tv_usec = 0;

		// oczekiwanie na dane lub timedout
		if ((res = select(max_fd+1, &rfds, NULL, NULL, &tv)) == -1) {
			if (errno == EINTR) continue;
			action(LOG_ERR, "select: %s", strerror(errno));
			continue;
		}

		// timedout
		if (res == 0) {
			res = LINE_CLOSED_TIMEOUT;
			break;
		}

		// dane od klienta
		if (FD_ISSET(data->client, &rfds)) {
			if (data->bdat_togo) {
				if ((res = bdat_chunk(data, 1)) == LINE_CLOSED) break;
			}

			// nie moze byc w "else", bo po bdat_chunk moga zostac jakies dane w buforze
			if (!data->bdat_togo) {
				assert(data->cli_size < buffer_size);
				res = fdgetline(data->client, data->cli_buf, buffer_size, &data->cli_size,
					&client_callback, data, &data->cli_rx);
				if (res == LINE_CLOSED) break;

				// cos moglo zostac w budorze
				// if (res == LINE_BINARY?) ?
				if (data->cli_size && data->bdat_togo) {
					if ((res = bdat_chunk(data, 0)) == LINE_CLOSED) break;
				}
			}

			SHARED_CONN_STATUS(cli_rx, data->cli_rx);
		}

		// dane z serwera
		if (FD_ISSET(data->server, &rfds)) {
			assert(data->srv_size < buffer_size);
			res = fdgetline(data->server, data->srv_buf, buffer_size, &data->srv_size,
				&server_callback, data, &data->srv_rx);
			
			SHARED_CONN_STATUS(srv_rx, data->srv_rx);
			
			if (res == LINE_CLOSED) {
				res = LINE_CLOSED_SERVER;
				break;
			}
		}
	}

	// porzadki

	SAFE_CLOSE(data->client);
	SAFE_CLOSE(data->server);

	if (!data->transaction) SHARED_STATS_INC(requests_empty);

	spool_remove(data);
	cleanup();

	action(LOG_NOTICE|LOG_ALWAYS, "CLOSE by=%s, rcv=%d/%d, trns=%d, rcpts=%d, time=%ld, host=%s, ident=%s",
		(res == LINE_CLOSED_TIMEOUT) ? "timeout" : ((res == LINE_CLOSED_SERVER) ? "server" : "client"),
		data->cli_rx, data->srv_rx, data->transaction, data->rcpts_total, time(NULL)-data->time,
		data->origin_str, data->ident);
	exit(0);
} /* connection() */


void dump_ver(int verbose)
{
	printf("SMTP Transparent AV proxy\n");
	CONF_SS2("version", VERSION);
	CONF_SS2("build date", __DATE__ " " __TIME__);

	if (!verbose) return;

	printf("Defines:\n");
#ifdef DEFAULT_CONFIG_FILE
	CONF_S2(DEFAULT_CONFIG_FILE);
#endif
	CONF_S2(HEADER);
	CONF_S2(SCANNER);
	CONF_D2(PIPELINE_QUEUE);
	CONF_D2(EMAIL_LENGTH);
#ifdef USE_SHARED_MEM
	CONF_SS2("USE_SHARED_MEM", "yes");
#else
	CONF_SS2("USE_SHARED_MEM", "no");
#endif
#ifdef USE_REGEX
	CONF_SS2("USE_REGEX", "yes");
#else
	CONF_SS2("USE_REGEX", "no");
#endif
#ifdef USE_PGSQL
	CONF_SS2("USE_PGSQL", "yes");
#else
	CONF_SS2("USE_PGSQL", "no");
#endif
#ifdef FILTER_CHUNKING
	CONF_SS2("FILTER_CHUNKING", "yes");
#else
	CONF_SS2("FILTER_CHUNKING", "no");
#endif
	printf("Extra:\n");
	CONF_D2(sizeof(struct proc_data));
} /* dump_ver() */

void dump_state()
{
	FILE *f;
	struct conn_info *ci;
	int i, td;
	int d, h, m, s;
	time_t now;
#ifdef USE_SHARED_MEM
	double speed;
#endif


	f = fopen(dumpfile, "w");
	if (!f) {
		action(LOG_ERR, "can't open dump file [%s]: %s", dumpfile, strerror(errno));
		return;
	}

	now = time(NULL);
	s = now - stats->started; // elapsed

	d = s / (3600*24);
	s %= (3600*24);
	h = s / 3600;
	s %= 3600;
	m = s / 60;
	s %= 60;

#ifndef USE_SHARED_MEM
	stats->rejects_ident =
		stats->viruses =
		stats->spams =
		stats->requests_empty = -1;
#endif
	
	fprintf(f, "Start time:   %s\n", time2str(stats->started));
	fprintf(f, "Restart time: %s\n", time2str(stats->restarted));
	fprintf(f, "Uptime:   %dd %dh %dm %ds\n", d, h, m, s);
	fprintf(f, "Found:    %d/%d (viruses/spam)\n", stats->viruses, stats->spams);
	fprintf(f, "Children: %d/%d (current/max)\n", children, stats->max_children);
	fprintf(f, "Requests: %d/%d/%d (total/direct/empty)\n",
		stats->requests, stats->requests_direct, stats->requests_empty);
	fprintf(f, "Rejects:  %d/%d/%d/%d (host/ident/lock/other)\n",
		stats->rejects_host, stats->rejects_ident, stats->rejects_lock, stats->rejects);

	fprintf(f, "\n");
	fprintf(f, "%4s %-5s %-8s %-5s %-15s %-15s %4s %8s %8s %7s %s\n", "slot", "pid", "state", "time", "source", "target", "trns", "cli_rx", "srv_rx", "kbps", "ident");
	for (i=0; i<max_connections_real; i++) {
		ci = &connections[i];
		if (!ci->pid) continue;

		td = now - ci->start_time;
#ifdef USE_SHARED_MEM
		if (td) {
			speed = ((float) ci->cli_rx)/(td*1024/8);
		} else {
			speed = 0;
		}
#endif
		
		fprintf(f, "%4d %-5d ", i, ci->pid);
#ifdef USE_SHARED_MEM
		fprintf(f, "%-8s", conn_states[ci->state]);
#else
		fprintf(f, "%-8s", "-");
#endif
		fprintf(f, " %02d:%02d", td / 60, td % 60);
		fprintf(f, " %-15s ", inet_ntoa(UINT32_TO_SIN(ci->src)));
#ifdef USE_SHARED_MEM
		fprintf(f, "%-15s %4d %8d %8d %7.1f", inet_ntoa(UINT32_TO_SIN(ci->dst)), ci->transaction, ci->cli_rx, ci->srv_rx, speed);
		fprintf(f, " %s", (ci->ident_ok) ? ci->ident : "");
#else
		fprintf(f, "%-15s %4s %8s %8s %6s", "-", "-", "-", "-", "-");
#endif

		fprintf(f, "\n");
	}

	fclose(f);
} /* dump_state() */


int setup()
{
	// max_connections defined during startup only,  this setting is ignored during reload
	max_connections_real = max_connections;

#ifdef USE_SHARED_MEM
	connections = shmalloc(sizeof(struct conn_info)*max_connections_real, &conn_shmid);
	if (connections == NULL) return -1;

	stats = shmalloc(sizeof(struct stat_info), &stat_shmid);
	if (stats == NULL) return -1;
#else
	connections = malloc(sizeof(struct conn_info) * max_connections_real);
	if (connections == NULL) {
		action(LOG_CRIT, "connections malloc failed: %s", strerror(errno));
		return -1;
	}

	stats = malloc(sizeof(struct stat_info));
	if (stats == NULL) {
		action(LOG_CRIT, "stats malloc failed: %s", strerror(errno));
		return -1;
	}
#endif

	return 0;
} /* setup */


/*
 * 	post-reconfiguration, called:
 * 	1. after daemonize() on start
 * 	2. after reading configuration on HUP
*/

void post_config()
{
	spool_max_size = (scan_max_size > spam_max_size) ? scan_max_size : spam_max_size;

	if (!EMPTY_STRING(locale)) {
		setlocale(LC_MESSAGES, locale);
//		action(LOG_DEBUG, "Changed locale to %s ['Success'=>'%s']", locale, strerror(0));
	}
} /* post_config() */


int start()
{
	int res, i, host_count, slot;
	int client, proxy;
	struct timeval tv;
	struct sockaddr_in origin;
	socklen_t originlen;
	double cur_load;
	struct proc_data data;


	memset(connections, 0, sizeof(struct conn_info) * max_connections_real);
	memset(stats, 0, sizeof(struct stat_info));

	stats->started = time(NULL);
	proxy = -1;

restart:
	stats->restarted = time(NULL);

	if ((res = setup_signal()) < 0) {
		action(LOG_CRIT, "setup_signal: %s", strerror(errno));
		cleanup();
		exit(1);
	}

	if (proxy == -1) {
		proxy = setup_listen(bind_address, port, connect_queue);
		if (proxy == -1) {
			cleanup();
			exit(1);
		}
		
		action(LOG_INFO, "SMTP-Proxy %s listening on %s:%d [queue: %d]", VERSION, bind_address, port, connect_queue);
	}

	for (;;) {
		if (force_finish) {
			force_finish = 0;
			break;
		}
		if (child_died) {
			// nie ma wyscigow, bo child_reaper obsluzy wszystkie zakonczone procesy
			child_died = 0;
			child_reaper();
		}
		if (force_reconfig) {
			force_reconfig = 0;
			action(LOG_INFO, "Reloading configuration...");
			if (read_config(config_options, config_file) != 0) {
				action(LOG_INFO, "Error parsing configuration file...");
				// inconsistent configuration 
				break;
			}

			post_config();
			if ((port >= 1024) || (getuid() == 0)) {
				SAFE_CLOSE(proxy);
				proxy = -1;
			} else {
				action(LOG_NOTICE, "NOTICE: listening socket won't be reopened...");
			}

			goto restart;
		}
		if (force_dump) {
			force_dump = 0;
			dump_state();
		}

		originlen = sizeof(origin);
		client = accept(proxy, (struct sockaddr*) &origin, &originlen);
		if (client == -1) {
			if (errno == EINTR) continue;
			action(LOG_CRIT, "accept error: %s", strerror(errno));
			break;
		}

		stats->requests++;

		/* fast DoS checks */
		if (children >= max_connections_real) {
			action(LOG_INFO, "Rejecting (%d) connection from %s:%d", children, inet_ntoa(origin.sin_addr), ntohs(origin.sin_port));
			fdprintf(client, "%03d %s %s\n", CONN_REJ_CODE, proxy_name, msg_max_reached);
			SAFE_CLOSE(client);
			stats->rejects++;
			continue;
		}

		if (is_load_above(max_load, &cur_load) == 1) {
			action(LOG_INFO, "Rejecting connection from %s:%d, load=%lf", inet_ntoa(origin.sin_addr), ntohs(origin.sin_port), cur_load);
			fdprintf(client, "%03d %s %s\n", CONN_REJ_CODE, proxy_name, msg_system_load);
			SAFE_CLOSE(client);
			stats->rejects++;
			continue;
		}

		slot = -1;
		host_count = 1;
		for (i=0; i<max_connections_real; i++) {
			if (connections[i].pid == 0) {
				slot = i;
			} else {
				if (connections[i].src == SIN_TO_UINT32(origin.sin_addr)) host_count++;
			}
		}

		if (host_count > max_per_host) {
			action(LOG_INFO, "Rejecting (host %d) connection from %s:%d", host_count, inet_ntoa(origin.sin_addr), ntohs(origin.sin_port));
			fdprintf(client, "%03d %s %s\n", CONN_REJ_CODE, proxy_name, msg_max_per_host);
			SAFE_CLOSE(client);
			stats->rejects_host++;
			continue;
		}

		if (slot == -1) {
			action(LOG_ALERT, "!BUG! No free slot, but max_connections_real=%d, rejecting connection from %s:%d", max_connections_real, inet_ntoa(origin.sin_addr), ntohs(origin.sin_port));
			fdprintf(client, "%03d %s %s\n", CONN_REJ_CODE, proxy_name, msg_temp_unavail);
			SAFE_CLOSE(client);
			stats->rejects++;
			continue;
		}

		context_init_1(&data, client, origin);

		if (lock_duration) {
			if (lockfile_present(&data)) {
				fdprintf(client, "%03d %s %s\n", CONN_REJ_CODE, proxy_name, msg_virus_locked);
				SAFE_CLOSE(client);
				stats->rejects_lock++;
				continue;
			}
		}

//		connections[slot].pid = -1;
		connections[slot].src = SIN_TO_UINT32(origin.sin_addr);
		connections[slot].start_time = time(NULL);
#ifdef USE_SHARED_MEM
		connections[slot].ident_ok = 
			connections[slot].dst =
			connections[slot].cli_rx =
			connections[slot].srv_rx =
			connections[slot].transaction = 0;
//		connections[slot].ident[]
		connections[slot].state = CONN_START;
#endif


		/* fork & service request */
#ifndef NO_FORK
		res = fork();
#else
		res = 0;
#endif
		if (res < 0) {
			action(LOG_CRIT, "fork error: %s", strerror(errno));
			break;
		}
		if (res) {	// parent
			connections[slot].pid = res;
			SAFE_CLOSE(client);
			children++;
			if (children > stats->max_children) stats->max_children = children;
			continue;
		}

		SAFE_CLOSE(proxy);

		i_am_a_child = 1;
		child_slot = slot;
		children++;
		connection(&data);
		cleanup();
		exit(0);
	}

	SAFE_CLOSE(proxy);

		
	if (children) {
		remove_pidfile();
		// hack => cleanup() nie bedzie probowal usunac ponownie
		if (pidfile) pidfile[0] = '\0';
		action(LOG_INFO, "Listening-socket closed, waiting for %d children(s) to finish", children);
	}

	while (children) {
		if (force_dump) {
			force_dump = 0;
			dump_state();
		}

		if (force_finish) break;

		if (child_died) {
			child_died = 0;
			child_reaper();
			if (!children) break;
		}
		
		tv.tv_sec = 3;
		tv.tv_usec = 0;
		if (select(0, NULL, NULL, NULL, &tv) == -1 && errno != EINTR) {
			action(LOG_ERR, "select: %s", strerror(errno));
			break;
		}
	}

	cleanup();
	action(LOG_INFO, "Children finished, exiting");
	return 0;
} /* start() */

int cat(char *fn)
{
	char buf[4096];
	int size, res, pos, fd;

	if ((fd = open(fn, O_RDONLY)) == -1) {
		fprintf(stderr, "open(%s) error: %s\n", fn, strerror(errno));
		return -1;
	}
	
	for (;;) {
		if ((size = read(fd, buf, sizeof(buf))) == -1) {
			if (errno == EINTR) continue;

			fprintf(stderr, "read() error: %s\n", strerror(errno));
			SAFE_CLOSE(fd);
			return -1;
		}

		if (size == 0) break;

		for (pos=0; size>0; size-=res, pos+=res) {
			if ((res = fwrite(buf+pos, 1, size, stdout)) == -1) {
				if (errno == EINTR) continue;

				fprintf(stderr, "cat:write() error: %s\n", strerror(errno));
				SAFE_CLOSE(fd);
				return -1;
			}
		}
	}

	SAFE_CLOSE(fd);
	return 0;
} /* cat() */


char* chrootize_path(char *path)
{
	static char buf[256];

	snprintf(buf, sizeof(buf), "%s%s%s", chroot_path, EMPTY_STRING(chroot_path) ? "" : "/./", path);
	TERMINATE_STRING(buf);
	
	return buf;
} /* chrootize_path () */

int signal_bg_process(int signum)
{
	char *real_pidfile;
	FILE *f;
	int bg_pid;

	real_pidfile = chrootize_path(pidfile);
	
	f = fopen(real_pidfile, "r");
	if (!f) {
		fprintf(stderr, "Cannot open .pid file [%s]: %s\n", real_pidfile, strerror(errno));
		return 1;
	}

	if (fscanf(f, "%d", &bg_pid) != 1) {
		fclose(f);
		fprintf(stderr, "Cannot parse .pid file [%s]\n", real_pidfile);
		return 1;
	}

	fclose(f);

//	fprintf(stderr, "Sending signal %d to process %d...\n", signal, bg_pid);
	if (kill(bg_pid, signum) != 0) {
		fprintf(stderr, "kill(%d, %d) error: %s\n", bg_pid, signum, strerror(errno));
		return 1;
	}

	return 0;
} /* signal_bg_process() */


void help()
{
	printf("SMTP Transparent AV proxy %s\n", VERSION);
	printf("Usage: smtp-gated [-f] [ -h | -s | -S | -r | -t | -T | -v | -V ] config_file\n");
	printf("	-f	run foreground\n");
	printf("	-h	this command reference\n");
	printf("	-s	prepare & show process status\n");
	printf("	-S	prepare & show process status file\n");
	printf("	-r	reload configuration\n");
	printf("	-t	syntax check & dump configuration (except messages)\n");
	printf("	-T	syntax check & dump configuration (including messages)\n");
	printf("	-v	show version\n");
	printf("	-V	show version & compiled-in options\n");
	printf("Signals:\n");
	printf("	HUP	reload configuration\n");
	printf("	USR1	dump statistics\n");
} /* help() */

// petla glowna

typedef enum {MODE_DEFAULT=0, MODE_MANY, MODE_H, MODE_V, MODE_VV, MODE_T, MODE_TT, MODE_R, MODE_S, MODE_SS} run_mode;


#define SET_MODE(x) if (mode == MODE_DEFAULT) mode = x; else mode = MODE_MANY;

int main(int argc, char* argv[])
{
	run_mode mode = MODE_DEFAULT;
	int i, res;


	for (i=1; i<argc; i++) {
		if ((strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "--help") == 0)) {
			SET_MODE(MODE_H);
		} else if (strcmp(argv[i], "-v") == 0) {
			SET_MODE(MODE_V);
		} else if (strcmp(argv[i], "-V") == 0) {
			SET_MODE(MODE_VV);
		} else if (strcmp(argv[i], "-t") == 0) {
			SET_MODE(MODE_T);
		} else if (strcmp(argv[i], "-T") == 0) {
			SET_MODE(MODE_TT);
		} else if (strcmp(argv[i], "-r") == 0) {
			SET_MODE(MODE_R);
		} else if (strcmp(argv[i], "-s") == 0) {
			SET_MODE(MODE_S);
		} else if (strcmp(argv[i], "-S") == 0) {
			SET_MODE(MODE_SS);
		} else if (strcmp(argv[i], "-f") == 0) {
			foreground++;
		} else {
			if (config_file) {
				fprintf(stderr, "You can supply only one configuration file!\n");
				exit(1);
			}

			config_file = argv[i];
		}
	}

	if (mode == MODE_MANY) {
		fprintf(stderr, "You can specify only one mode\n");
		exit(3);
	}

#ifdef DEFAULT_CONFIG_FILE
	if (!config_file) config_file = DEFAULT_CONFIG_FILE;
#endif

	foreground++;
	if (config_file) {
		if (read_config(config_options, config_file) != 0) {
			if (config_file[0] == '-') fprintf(stderr, "Try %s -h for help\n", argv[0]);
			exit(2);
		}
	} else {
		if ((res = config_set_defaults(config_options)) != 0) {
			fprintf(stderr, "!BUG! invalid default (hardcoded) config [%d]\n", res);
			exit(10);
		}
	}
	foreground--;

	switch (mode) {
		case MODE_DEFAULT:
			break;
		case MODE_H:
			help();
			exit(10);
		case MODE_V:
		case MODE_VV:
			dump_ver(mode == MODE_VV);
			exit(0);
		case MODE_T:
		case MODE_TT:
			dump_config(config_options, (mode == MODE_TT) ? NULL : "msg_");
			exit(0);
		case MODE_S:
		case MODE_SS:
			if (signal_bg_process(SIGUSR1) != 0) exit(1);
			sleep(1);
			if (mode == MODE_SS) {
				printf("%s\n", chrootize_path(dumpfile));
			} else {
				i = cat(chrootize_path(dumpfile));
				if (i) exit(1);
			}
			exit(0);
		case MODE_R:
			if (signal_bg_process(SIGHUP) != 0) exit(1);
			exit(0);
		default:
			fprintf(stderr, "unknown exec mode [%d]\n", mode);
			exit(3);
	}
	
	if (!config_file) {
		fprintf(stderr, "Config file not supplied.\n");
		exit(1);
	}

	if (daemonize() != 0) exit(1);
	post_config();

	if (setup() != 0) {
		cleanup();
		exit(1);
	}

	start();
	exit(0);
} /* main() */

// : vim: 

